from datetime import datetime, timezone, timedelta

import requests

from shipments.structs import PackageInfo, PackageEvent, StatusCategory, Address


class Carrier:
    DISPLAYNAME = 'N/A'

    @classmethod
    def identify(cls, code):
        pass

    @classmethod
    def get_requirements(cls, code):
        return []

    def get_info(self, code, extra):
        pass


class PostNL(Carrier):
    DISPLAYNAME = 'PostNL'

    @classmethod
    def identify(cls, code):
        if code.startswith('3S'):
            if len(code) == 13:
                return True
        if code.startswith('KG'):
            if len(code) == 8:
                return True
        return False

    @classmethod
    def get_requirements(cls, code):
        return [
            ('country', 'Destination country', 'country-alpha2:NL'),
            ('postalcode', 'Postal code', 'text')
        ]

    def get_info(self, code, extra):
        if not 'country' in extra:
            raise ValueError("Missing 'country' in extra data")
        if not 'postalcode' in extra:
            raise ValueError("Missing 'postalcode' in extra data")
        country = extra['country'].upper()
        postalcode = extra['postalcode'].upper().replace(' ', '').replace('-', '')
        url = f'https://jouw.postnl.nl/track-and-trace/api/trackAndTrace/{code}-{country}-{postalcode}?language=en'
        response = requests.get(url)
        data = response.json()

        data = data['colli'][code]

        result = PackageInfo(code, 'PostNL')
        result.carrier_name = 'PostNL'
        result.status = 'Label created'
        result.status_category = StatusCategory.LABEL_CREATED
        result.tracking_url = f'https://jouw.postnl.nl/track-and-trace/{code}-{country}-{postalcode}?language=en'
        if data['isDelivered']:
            result.status = 'Delivered'
            result.status_category = StatusCategory.DELIVERED
        result.description = data['description']

        for event in data['observations']:
            stamp = datetime.fromisoformat(event['observationDate'])
            e = PackageEvent(stamp, '', event['description'])
            result.events.append(e)

        r = Address()
        r.name = data['recipient']['names']['companyName']
        r.attention = data['recipient']['names']['personName']
        r.address1 = data['recipient']['address']['street'] + ' ' + data['recipient']['address']['houseNumber']
        if data['recipient']['address']['houseNumberSuffix']:
            r.address1 += ' ' + data['recipient']['address']['houseNumberSuffix']
        r.postalcode = data['recipient']['address']['postalCode']
        r.city = data['recipient']['address']['town']
        r.country = data['recipient']['address']['country']
        result.destination = r

        s = Address()
        s.name = data['sender']['names']['companyName']
        s.attention = data['sender']['names']['personName']
        s.address1 = data['sender']['address']['street'] + ' ' + data['sender']['address']['houseNumber']
        if data['sender']['address']['houseNumberSuffix']:
            s.address1 += ' ' + data['sender']['address']['houseNumberSuffix']

        s.postalcode = data['sender']['address']['postalCode']
        s.city = data['sender']['address']['town']
        s.country = data['sender']['address']['country']
        result.origin = s

        try:
            result.weight = int(data['details']['dimensions']['weight'])
        except:
            pass
        return result


class RussianPost(Carrier):
    DISPLAYNAME = 'Russian Post'

    @classmethod
    def identify(cls, code):
        if len(code) != 14:
            return False

        a = 0
        b = 0
        x = True
        for char in code[:-1]:
            if not char.isdigit():
                return False
            num = int(char)

            if x:
                a += int(num) * 3
            else:
                b += int(num)
            x = not x

        s = (a + b) % 10
        calculated = 10 - s
        if code[-1].isdigit():
            check = int(code[-1])
        else:
            return False

        return calculated == check

    def get_info(self, code, extra):
        url = 'https://www.pochta.ru/tracking?p_p_id=trackingPortlet_WAR_portalportlet&p_p_lifecycle=2&p_p_resource_id=tracking.get-by-barcodes'
        response = requests.post(url, data={'barcodes': code})
        payload = response.json()['response'][0]['trackingItem']

        result = PackageInfo(code, self.DISPLAYNAME)
        result.carrier_name = self.DISPLAYNAME
        result.status = payload['commonStatus']
        if payload['globalStatus'] == 'ARCHIVED':
            result.status_category = StatusCategory.DELIVERED
        elif payload['globalStatus'] == 'REGISTERED':
            result.status_category = StatusCategory.LABEL_CREATED
        else:
            result.status_category = StatusCategory.IN_TRANSIT

        result.description = payload['title']
        result.delivery_note = payload['mailType'] + ' (' + payload['mailCtgText'] + ')'
        result.weight = payload['weight']

        for event in payload['trackingHistoryItemList']:
            # Timestamp is a full ISO 8601 with timezones
            stamp = datetime.fromisoformat(event['date'][0:10])
            e = PackageEvent(stamp, event['description'], event['humanStatus'])
            result.events.append(e)

        if result.status == "":
            result.status = result.events[0].label

        r = Address()
        r.name = payload['recipient']
        r.city = payload['destinationCityName']
        r.country = payload['destinationCountryName']
        r.postalcode = payload['indexTo']
        result.destination = r

        s = Address()
        s.name = payload['sender']
        s.city = payload['originCityName']
        s.country = payload['originCountryName']
        s.postalcode = payload['indexFrom']
        result.origin = s

        return result


class UPS(Carrier):
    DISPLAYNAME = 'UPS'

    @classmethod
    def identify(cls, code):
        if not code.startswith('1Z'):
            return False

        a = 0
        b = 0
        x = True
        for char in code[2:-1]:
            if char.isdigit():
                num = int(char)
            else:
                num = (ord(char) - 63) % 10
            if x:
                a += int(num)
            else:
                b += int(num) * 2
            x = not x
        s = (a + b) % 10
        calculated = 10 - s
        if code[-1].isdigit():
            check = int(code[-1])
        else:
            check = (ord(code[-1]) - 63) % 10

        return calculated == check

    def get_info(self, code, extra):
        session = requests.Session()
        user_url = 'https://www.ups.com/track'
        response = session.get(user_url)
        token = response.cookies['X-XSRF-TOKEN-ST']
        url = 'https://www.ups.com/track/api/Track/GetStatus'
        request_data = {
            'Locale': 'en_US',
            'Requester': 'st/trackdetails',
            'TrackingNumber': [code]
        }
        response = session.post(url, json=request_data, headers={
            'Content-Type': 'Application/Json',
            'Referer': 'https://www.ups.com/track',
            'DNT': '1',
            'User-Agent': 'Mozilla/5.0',
            'X-Requested-With': 'XMLHttpRequest',
            "Accept": "application/json, text/plain, */*",
            'X-XSRF-TOKEN': token
        })
        payload = response.json()

        if payload['trackDetails'] is None:
            result = PackageInfo(code, 'UPS')
            result.status = payload['statusText']
            result.carrier_name = 'UPS'
            result.status_category = StatusCategory.ERROR
            return result

        details = payload['trackDetails'][0]
        result = PackageInfo(code, 'UPS')
        result.carrier_name = 'UPS'
        result.tracking_url = 'https://www.ups.com/track?loc=nl_NL&requester=DAN&tracknum=' + code
        result.status = details['packageStatus']
        result.status_category = StatusCategory.LABEL_CREATED
        if details['progressBarType'] == 'InTransit':
            result.status_category = StatusCategory.IN_TRANSIT
        elif details['progressBarType'] == 'Delivered':
            result.status_category = StatusCategory.DELIVERED

        for event in details['shipmentProgressActivities']:
            better_time = event['time'].replace('.M.', 'M').zfill(8)
            dt = event['date'] + ' ' + better_time
            pdt = datetime.strptime(dt, '%m/%d/%Y %I:%M %p')
            label = 'No details'
            if event['actCode'] == 'MP':
                label = 'Label created'
            elif event['actCode'] == 'OF':
                label = 'Out for delivery'
            if event['milestoneName'] is not None and 'name' in event['milestoneName']:
                label = event['milestoneName']['name']
            elif 'activityScan' in event and event['activityScan'] is not None:
                label = event['activityScan']
            e = PackageEvent(pdt, event['location'], label)
            result.events.append(e)

        if details['shipToAddress'] is not None:
            a = details['shipToAddress']
            sa = Address()
            sa.name = a['companyName']
            sa.attention = a['attentionName']
            sa.address1 = a['streetAddress1']
            sa.address2 = a['streetAddress2']
            sa.address3 = a['streetAddress3']
            sa.postalcode = a['zipCode']
            sa.city = a['city']
            sa.state = (a['state'] + ' ' + a['province']).strip()
            sa.country = a['country']
            result.destination = sa

        if details['additionalInformation'] is not None:
            ai = details['additionalInformation']
            if 'weight' in ai:
                weight = ai['weight']
                unit = ai['weightUnit']
                if unit == 'KGS':  # WTF ups, that's not how you write it
                    result.weight = int(float(weight) * 1000)
            if 'serviceInformation' in ai and ai['serviceInformation'] is not None:
                result.service = ai['serviceInformation']['serviceName'].replace('&#174;', '')

        if 'leftAt' in details and details['leftAt'] is not None and details['leftAt'] != '':
            result.delivery_note = details['leftAt']
        return result


class InPost(Carrier):
    DISPLAYNAME = 'InPost'

    STATUS_MAPPING = {
        StatusCategory.LABEL_CREATED: ['created', 'offers_prepared', 'offer_selected', 'confirmed',
                                       'dispatched_by_sender_to_pok'],
        StatusCategory.IN_TRANSIT: ['dispatched_by_sender', 'collected_from_sender', 'taken_by_courier',
                                    'adopted_at_source_branch', 'sent_from_source_branch', 'adopted_at_sorting_center',
                                    'sent_from_sorting_center', 'adopted_at_target_branch', 'out_for_delivery',
                                    'out_for_delivery_to_address', 'pickup_reminder_sent_address',
                                    'taken_by_courier_from_pok', 'ready_to_pickup_from_branch', 'delay_in_delivery',
                                    'redirect_to_box', 'canceled_redirect_to_box', 'readdressed',
                                    'stack_in_customer_service_point', 'unstack_from_customer_service_point',
                                    'unstack_from_box_machine', 'stack_parcel_in_box_machine_pickup_time_expired',
                                    'stack_parcel_pickup_time_expired'],
        StatusCategory.DELIVERED: ['ready_to_pickup_from_pok', 'ready_to_pickup_from_pok_registered', 'ready_to_pickup',
                                   'pickup_reminder_sent', 'delivered', 'return_pickup_confirmation_to_sender',
                                   'courier_avizo_in_customer_service_point', 'stack_in_box_machine'],
        StatusCategory.ERROR: ['oversized', 'pickup_time_expired', 'avizo', 'claimed', 'returned_to_sender', 'canceled',
                               'other', 'rejected_by_receiver', 'undelivered_wrong_address',
                               'undelivered_incomplete_address', 'undelivered_unknown_receiver',
                               'undelivered_cod_cash_receiver', 'undelivered', 'undelivered_no_mailbox',
                               'undelivered_not_live_address', 'undelivered_lack_of_access_letterbox', 'missing',
                               'taken_by_courier_from_customer_service_point']
    }

    @classmethod
    def identify(cls, code):
        if not len(code) == 24:
            return False
        return True

    def get_info(self, code, extra):
        url = 'https://inpost.pl/shipx-proxy/'
        params = {'number': code}
        headers = {'X-Requested-With': 'XMLHttpRequest'}
        response = requests.get(url, params=params, headers=headers)
        payload = response.json()
        result = PackageInfo(code, 'InPost')
        result.carrier_name = self.DISPLAYNAME
        if 'error' in payload and payload['error'] is not None:
            result.status = payload['error']
            result.status_category = StatusCategory.ERROR
            return result
        result.status = " ".join(payload['status'].split('_')).capitalize()

        for category, statuses in self.STATUS_MAPPING.items():
            if payload['status'] in statuses:
                result.status_category = category

        result.delivery_note = None
        result.description = None
        result.service = payload['service']
        if result.service in ['inpost_locker_standard', 'inpost_locker_allegro', 'inpost_locker_allegro_smart']:
            target = payload['custom_attributes']['target_machine_detail']
            a = Address()
            a.name = f'{target["name"]} - {target["location_description"]}'
            a.address1 = target['address']['line1']
            (a.postalcode, a.city) = target['address']['line2'].split(' ')
            a.country = 'Poland'
            result.destination = a

            if payload['custom_attributes'].get('dropoff_machine_detail'):
                dropoff = payload['custom_attributes']['dropoff_machine_detail']
                a = Address()
                a.name = f'{dropoff["name"]} - {dropoff["location_description"]}'
                a.address1 = dropoff['address']['line1']
                (a.postalcode, a.city) = dropoff['address']['line2'].split(' ')
                a.country = 'Poland'
                result.origin = a

        for event in payload['tracking_details']:
            stamp = datetime.fromisoformat(event['datetime'])
            e = PackageEvent(stamp, '', " ".join(event['status'].split('_')).capitalize())
            result.events.append(e)

        return result


class DHL(Carrier):
    DISPLAYNAME = 'DHL'

    def _addr(self, payload):
        res = Address()
        if 'countryCode' in payload['address']:
            res.country = payload['address']['countryCode']
        return res

    def get_requirements(cls, code):
        return [
            ('apikey', 'DHL API key', 'text'),
        ]

    def get_info(self, code, extra):
        url = 'https://api-eu.dhl.com/track/shipments'
        headers = {
            'Accept': 'application/json',
            'DHL-API-Key': extra['apikey'],
        }
        payload = {
            'trackingNumber': code
        }
        response = requests.get(url, params=payload, headers=headers)
        payload = response.json()
        result = PackageInfo(code, 'DHL')
        result.carrier_name = self.DISPLAYNAME
        result.status_category = StatusCategory.ERROR
        result.tracking_url = 'https://www.dhl.com/gb-en/home/tracking/tracking-parcel.html?submit=1&tracking-id=' + code

        if len(payload['shipments']):
            payload = payload['shipments'][0]
        else:
            result.status = payload['']
            return result

        result.status_category = StatusCategory.LABEL_CREATED
        result.status = payload['status']['status']
        if payload['status']['statusCode'] == 'pre-transit':
            result.status_category = StatusCategory.LABEL_CREATED
        elif payload['status']['statusCode'] == 'transit':
            result.status_category = StatusCategory.IN_TRANSIT

        if 'weight' in payload['details'] and payload['details']['weight'] is not None:
            w = payload['details']['weight']
            if w['unitText'] == 'kg':
                result.weight = w['value'] * 1000
            else:
                print(f"Unknown unit '{w['unitText']}'")

        for event in payload['events']:
            stamp = datetime.fromisoformat(event['timestamp'])
            location = ''
            if 'location' in event:
                if len(event['location']['address']) == 1 and 'addressLocality' in event['location']['address']:
                    location = event['location']['address']['addressLocality']
            e = PackageEvent(stamp, location, event['status'])
            result.events.append(e)

        if 'address' in payload['origin']:
            result.origin = self._addr(payload['origin'])

        if 'address' in payload['destination']:
            result.destination = self._addr(payload['destination'])

        return result


class Track4PX(Carrier):
    DISPLAYNAME = '4PX'

    @classmethod
    def identify(cls, code):
        if len(code) != 20:
            return False
        if not code.startswith('H'):
            return False
        if not code[1:].isnumeric():
            return False
        return True

    def get_info(self, code, extra):
        url = 'https://track.4px.com/track/v2/front/listTrack'
        data = {
            'serveCodes': [code],
            'language': 'en-us',
        }
        response = requests.post(url, json=data)
        payload = response.json()
        result = PackageInfo(code, 'Track4PX')
        result.carrier_name = self.DISPLAYNAME
        result.status_category = StatusCategory.ERROR
        result.tracking_url = 'https://track.4px.com/#/result/0/' + code

        if 'data' in payload and len(payload['data']):
            payload = payload['data'][0]
        else:
            result.status = payload['message']
            return result

        if payload['status'] == 0:
            result.status_category = StatusCategory.ERROR
            result.status = 'error'
        elif payload['status'] == 1:
            result.status_category = StatusCategory.IN_TRANSIT
            result.status = 'In transit'
        else:
            result.status_category = StatusCategory.DELIVERED
            result.status = 'Delivered'

        for event in payload['tracks']:
            _, tz = event['tkTimezone'].split('+')
            tz_hour, tz_min = tz.split(':')
            stamp = datetime.fromtimestamp(event['tkDate'] / 1000,
                                           tz=timezone(timedelta(hours=int(tz_hour, 10), minutes=int(tz_min, 10))))
            e = PackageEvent(stamp, event['tkLocation'], event['tkDesc'])
            result.events.append(e)

        result.origin = Address()
        result.origin.country = payload['ctStartName']
        result.destination = Address()
        result.destination.country = payload['ctEndName']

        return result
